@@ -23,7 +23,17 @@ from featurestore.core.data_source_wrappers import (
     SparkDataFrame,
     SparkPipeline,
 )
-from featurestore.core.schema import Column, Schema
+from featurestore.core.schema import FeatureSchema, Schema
+from featurestore.core.user_credentials import (
+    AWSCredentials,
+    AzureKeyCredentials,
+    AzurePrincipalCredentials,
+    AzureSasCredentials,
+    DriverlessAICredentials,
+    PostgresCredentials,
+    SnowflakeCredentials,
+    TeradataCredentials,
+)
 
 
 def __get_version():
@@ -39,10 +49,14 @@ def __get_version():
 __version__ = __get_version()
 __all__ = [
     "Client",
-    "Column",
+    "FeatureSchema",
+    "Schema",
     "CSVFile",
+    "CSVFolder",
     "JSONFile",
+    "JSONFolder",
     "ParquetFile",
+    "ParquetFolder",
     "SnowflakeTable",
     "JdbcTable",
     "PartitionOptions",
@@ -56,4 +70,12 @@ __all__ = [
     "Schema",
     "SparkDataFrame",
     "ClientConfig",
+    "AzureKeyCredentials",
+    "AzureSasCredentials",
+    "AzurePrincipalCredentials",
+    "AWSCredentials",
+    "TeradataCredentials",
+    "SnowflakeCredentials",
+    "PostgresCredentials",
+    "DriverlessAICredentials",
 ]
