@@ -1,12 +1,10 @@
 from typing import List
 
-from google.protobuf.empty_pb2 import Empty
-
 from .. import CoreService_pb2 as pb
 from ..entities.base_job import BaseJob
 from ..entities.extract_schema_job import ExtractSchemaJob
 from ..entities.ingest_job import IngestJob
-from ..entities.project import Project
+from ..entities.materialization_online_job import MaterializationOnlineJob
 from ..entities.retrieve_job import RetrieveJob
 
 
@@ -15,7 +13,7 @@ class Jobs:
         self._stub = stub
 
     @staticmethod
-    def create_job(stub, job_proto):
+    def _create_job(stub, job_proto):
         job_id = pb.JobId(job_id=job_proto.job_id)
         if job_proto.job_type == pb.JobType.Ingest:
             return IngestJob(stub, job_id)
@@ -23,16 +21,18 @@ class Jobs:
             return ExtractSchemaJob(stub, job_id)
         elif job_proto.job_type == pb.JobType.Retrieve:
             return RetrieveJob(stub, job_id)
+        elif job_proto.job_type == pb.JobType.MaterializationOnline:
+            return MaterializationOnlineJob(stub, job_id)
 
     def list(self, active=True, job_type=pb.JobType.Unknown) -> List[BaseJob]:
         request = pb.ListJobsRequest(active=active, job_type=job_type)
         resp = self._stub.ListJobs(request)
-        return [self.create_job(self._stub, job_proto) for job_proto in resp.jobs]
+        return [Jobs._create_job(self._stub, job_proto) for job_proto in resp.jobs]
 
     def get(self, job_id: str) -> BaseJob:
         request = pb.JobId(job_id=job_id)
         job_proto = self._stub.GetJob(request)
-        return self.create_job(self._stub, job_proto)
+        return Jobs._create_job(self._stub, job_proto)
 
     def __repr__(self):
         return "This class wraps together methods working with jobs"
