@@ -1,8 +1,6 @@
 import shutil
 import tempfile
 
-from google.protobuf.empty_pb2 import Empty
-
 import featurestore.core.CoreService_pb2 as pb
 import featurestore.core.FeatureSetSearch_pb2 as pb_search
 from featurestore.core.FeatureSetSearch_pb2 import (
@@ -48,11 +46,17 @@ class SparkDataFrame:
         self._write_path = None
 
     def _write_to_cache(self, stub):
-        write_info = stub.GetIngestWriteCredentials(Empty())
         from pyspark.sql import SparkSession  # Local import
 
         spark = SparkSession.builder.getOrCreate()
+
+        session_id = spark.conf.get("ai.h2o.featurestore.sessionId", "")
+        request = pb.IngestWriteCredentialRequest()
+        request.session_id = session_id
+        write_info = stub.GetIngestWriteCredentials(request)
+
         SparkUtils.configure_user_spark(spark)
+        spark.conf.set("ai.h2o.featurestore.sessionId", write_info.session_id)
         for k, v in write_info.options.items():
             spark.conf.set(k, v)
             if k.startswith("spark.hadoop."):
