@@ -1,10 +1,8 @@
-import json
 from abc import ABC, abstractmethod
 from concurrent.futures import ThreadPoolExecutor
 
 from .. import CoreService_pb2 as pb
 from .. import interactive_console
-from ..utils import Utils
 
 
 class BaseJob(ABC):
@@ -29,7 +27,7 @@ class BaseJob(ABC):
         return self.is_done()
 
     @abstractmethod
-    def _response_method(self):
+    def _response_method(self, job_id):
         raise NotImplementedError(
             "Method `_response_method` needs to be implemented by the child class"
         )
