@@ -1,6 +1,10 @@
+import json
 import time
 
 from .. import CoreService_pb2 as pb
+from .. import FeatureSetProtoApi_pb2 as FeatureSetApi
+from .. import OnlineApi_pb2 as OnlineApi
+from .. import TimeToLiveApi_pb2 as TimeToLiveApi
 from .. import interactive_console
 from ..collections.ingest_history import IngestHistory
 from ..credentials import CredentialsHelper
@@ -10,6 +14,7 @@ from ..retrieve_holder import RetrieveHolder
 from ..schema import Schema
 from ..utils import Utils
 from .ingest_job import IngestJob
+from .materialization_online_job import MaterializationOnlineJob
 from .user import User
 
 
@@ -19,6 +24,10 @@ class FeatureSet:
         fs.CopyFrom(feature_set)
         self._feature_set = fs
         self._stub = stub
+        self._feature_set_header = FeatureSetApi.FeatureSetHeader()
+        self._feature_set_header.project_id = self._feature_set.project_id
+        self._feature_set_header.feature_set_id = self._feature_set.id
+        self._feature_set_header.feature_set_version = self._feature_set.version
 
     @property
     def project(self):
@@ -28,10 +37,6 @@ class FeatureSet:
     def feature_set_name(self):
         return self._feature_set.feature_set_name
 
-    @property
-    def raw_data_location(self):
-        return self._feature_set.raw_data_location
-
     @property
     def version(self):
         return self._feature_set.version
@@ -44,17 +49,32 @@ class FeatureSet:
     def time_travel_column(self):
         return self._feature_set.time_travel_column
 
+    @property
+    def partition_by(self):
+        return self._feature_set.partition_by
+
     @property
     def time_travel_column_format(self):
         return self._feature_set.time_travel_column_format
 
     @property
     def feature_set_type(self):
-        return self._feature_set.feature_set_type
+        return FeatureSetApi.FeatureSetType.Name(self._feature_set.feature_set_type)
 
     @feature_set_type.setter
     def feature_set_type(self, value):
-        self._feature_set.feature_set_type = value
+        valid_values = FeatureSetApi.FeatureSetType.keys()
+        if value.upper() in valid_values:
+            update_request = pb.FeatureSetTypeUpdateRequest()
+            update_request.new_value = FeatureSetApi.FeatureSetType.Value(value.upper())
+            update_request.header.CopyFrom(self._feature_set_header)
+            self._stub.UpdateFeatureSetType(update_request)
+            self.refresh()
+        else:
+            raise Exception(
+                "Invalid feature set type. Supported values are: "
+                + ", ".join(map(str, valid_values))
+            )
 
     @property
     def description(self):
@@ -62,7 +82,11 @@ class FeatureSet:
 
     @description.setter
     def description(self, value):
-        self._feature_set.description = value
+        update_request = pb.FeatureSetStringFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetDescription(update_request)
+        self.refresh()
 
     @property
     def owner(self):
@@ -74,9 +98,11 @@ class FeatureSet:
         request.email = email
         response = self._stub.GetUserByMail(request)
         user = response.user
-        self._feature_set.owner.id = user.id
-        self._feature_set.owner.name = user.name
-        self._feature_set.owner.email = user.email
+        update_request = pb.FeatureSetUserFieldUpdateRequest()
+        update_request.new_value.CopyFrom(user)
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetOwner(update_request)
+        self.refresh()
 
     @property
     def author(self):
@@ -96,7 +122,11 @@ class FeatureSet:
 
     @application_name.setter
     def application_name(self, value):
-        self._feature_set.application_name = value
+        update_request = pb.FeatureSetStringFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetApplicationName(update_request)
+        self.refresh()
 
     @property
     def deprecated(self):
@@ -104,7 +134,11 @@ class FeatureSet:
 
     @deprecated.setter
     def deprecated(self, value):
-        self._feature_set.deprecated = value
+        update_request = pb.FeatureSetBooleanFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetDeprecated(update_request)
+        self.refresh()
 
     @property
     def deprecated_date(self):
@@ -120,7 +154,11 @@ class FeatureSet:
             raise ValueError(
                 "data_source_domains accepts only list of strings as a value"
             )
-        self._feature_set.data_source_domains[:] = value
+        update_request = pb.FeatureSetStringArrayFieldUpdateRequest()
+        update_request.new_value.extend(value)
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetDataSourceDomains(update_request)
+        self.refresh()
 
     @property
     def tags(self):
@@ -130,7 +168,11 @@ class FeatureSet:
     def tags(self, value):
         if not isinstance(value, list):
             raise ValueError("tags accepts only list of strings as a value")
-        self._feature_set.tags[:] = value
+        update_request = pb.FeatureSetStringArrayFieldUpdateRequest()
+        update_request.new_value.extend(value)
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetTags(update_request)
+        self.refresh()
 
     @property
     def process_interval(self):
@@ -138,15 +180,30 @@ class FeatureSet:
 
     @process_interval.setter
     def process_interval(self, value):
-        self._feature_set.process_interval = value
+        update_request = pb.FeatureSetIntFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetProcessInterval(update_request)
+        self.refresh()
 
     @property
     def process_interval_unit(self):
-        return self._feature_set.process_interval_unit
+        return pb.ProcessIntervalUnit.Name(self._feature_set.process_interval_unit)
 
     @process_interval_unit.setter
     def process_interval_unit(self, value):
-        self._feature_set.process_interval_unit = value
+        valid_units = pb.ProcessIntervalUnit.keys()
+        if value.upper() in valid_units:
+            update_request = pb.FeatureSetProcessIntervalUnitUpdateRequest()
+            update_request.new_value = pb.ProcessIntervalUnit.Value(value.upper())
+            update_request.header.CopyFrom(self._feature_set_header)
+            self._stub.UpdateFeatureSetProcessIntervalUnit(update_request)
+            self.refresh()
+        else:
+            raise Exception(
+                "Invalid process interval unit. Supported values are: "
+                + ", ".join(map(str, valid_units))
+            )
 
     @property
     def flow(self):
@@ -154,12 +211,17 @@ class FeatureSet:
 
     @flow.setter
     def flow(self, value):
-        self._feature_set.flow = value
+        update_request = pb.FeatureSetStringFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetFlow(update_request)
+        self.refresh()
 
     @property
     def features(self):
         return {
-            feature.name: Feature(feature) for feature in self._feature_set.features
+            feature.name: Feature(self._stub, self, feature, feature.name)
+            for feature in self._feature_set.features
         }
 
     @property
@@ -168,7 +230,11 @@ class FeatureSet:
 
     @primary_key.setter
     def primary_key(self, value):
-        self._feature_set.primary_key = value
+        update_request = pb.FeatureSetStringFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetPrimaryKey(update_request)
+        self.refresh()
 
     @property
     def secondary_key(self):
@@ -176,7 +242,11 @@ class FeatureSet:
 
     @secondary_key.setter
     def secondary_key(self, value):
-        self._feature_set.secondary_key = value
+        update_request = pb.FeatureSetStringFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetSecondaryKey(update_request)
+        self.refresh()
 
     @property
     def statistics(self):
@@ -184,15 +254,11 @@ class FeatureSet:
 
     @property
     def time_to_live(self):
-        return TimeToLive(self._feature_set)
+        return TimeToLive(self._stub, self)
 
     @property
     def special_data(self):
-        return SpecialData(self._feature_set)
-
-    @flow.setter
-    def flow(self, value):
-        self._feature_set.flow = value
+        return FeatureSetSpecialData(self._stub, self)
 
     @property
     def time_travel_scope(self):
@@ -204,7 +270,11 @@ class FeatureSet:
 
     @application_id.setter
     def application_id(self, value):
-        self._feature_set.application_id = value
+        update_request = pb.FeatureSetStringFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetApplicationId(update_request)
+        self.refresh()
 
     @property
     def feature_set_state(self):
@@ -212,7 +282,11 @@ class FeatureSet:
 
     @feature_set_state.setter
     def feature_set_state(self, value):
-        self._feature_set.feature_set_state = value
+        update_request = pb.FeatureSetStringFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetState(update_request)
+        self.refresh()
 
     @property
     def online(self):
@@ -224,7 +298,23 @@ class FeatureSet:
 
     @secret.setter
     def secret(self, value):
-        self._feature_set.secret = value
+        update_request = pb.FeatureSetBooleanFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetSecret(update_request)
+        self.refresh()
+
+    @property
+    def custom_data(self):
+        return self._feature_set.custom_data
+
+    @custom_data.setter
+    def custom_data(self, value):
+        update_request = pb.FeatureSetCustomDataUpdateRequest()
+        update_request.new_value.CopyFrom(value)
+        update_request.header.CopyFrom(self._feature_set_header)
+        self._stub.UpdateFeatureSetCustomData(update_request)
+        self.refresh()
 
     def create_new_version(self, schema=None, affected_features=None, reason=""):
         if schema is None and affected_features is None:
@@ -235,19 +325,21 @@ class FeatureSet:
         request.feature_set.CopyFrom(self._feature_set)
         request.reason = reason
         if schema:
-            request.schema.extend(schema.dump())
+            request.schema.extend(schema._to_proto_schema())
         else:
-            request.schema.extend(self.get_schema().dump())
+            request.schema.extend(self.get_schema()._to_proto_schema())
         if affected_features:
             request.affected_features.extend(affected_features)
         response = self._stub.CreateNewFeatureSetVersion(request)
         return FeatureSet(self._stub, response.feature_set)
 
-    def update_metadata(self):
-        request = pb.UpdateFeatureSetMetadataRequest()
-        request.feature_set.CopyFrom(self._feature_set)
-        response = self._stub.UpdateFeatureSetMetadata(request)
-        self._feature_set = response.feature_set
+    def refresh(self):
+        request = pb.GetFeatureSetsLastMinorForCurrentMajorRequest()
+        request.header.CopyFrom(self._feature_set_header)
+        self._feature_set = self._stub.GetFeatureSetsLastMinorForCurrentMajor(
+            request
+        ).feature_set
+        self._feature_set_header.feature_set_version = self._feature_set.version
         return self
 
     def delete(self, wait_for_completion=False):
@@ -299,7 +391,7 @@ class FeatureSet:
         request.feature_set.CopyFrom(self._feature_set)
         request.job_type = job_type
         resp = self._stub.ListJobs(request)
-        return [Jobs.create_job(self._stub, job_proto) for job_proto in resp.jobs]
+        return [Jobs._create_job(self._stub, job_proto) for job_proto in resp.jobs]
 
     def _add_permissions(self, user_emails, permission):
         request = pb.FeatureSetPermissionRequest()
@@ -323,6 +415,7 @@ class FeatureSet:
         new_version_on_schema_change=False,
         start_date_time=None,
         end_date_time=None,
+        credentials=None,
     ):
         from ..data_source_wrappers import get_raw_data_location
 
@@ -335,7 +428,7 @@ class FeatureSet:
         request.feature_set.CopyFrom(self._feature_set)
         request.new_version_on_schema_change = new_version_on_schema_change
         request.data_source.CopyFrom(data_source)
-        CredentialsHelper.set_credentials(request, data_source)
+        CredentialsHelper.set_credentials(request, data_source, credentials)
         if start_date_time is not None:
             request.start_date_time = start_date_time
         if end_date_time is not None:
@@ -350,23 +443,57 @@ class FeatureSet:
         new_version_on_schema_change=False,
         start_date_time=None,
         end_date_time=None,
+        credentials=None,
     ):
         job = self.ingest_async(
-            source, new_version_on_schema_change, start_date_time, end_date_time
+            source,
+            new_version_on_schema_change,
+            start_date_time,
+            end_date_time,
+            credentials,
         )
         while not job.is_done():
             job.show_progress()
             time.sleep(2)
+        job.show_progress()  # there is possibility that some progress was pushed before finishing job
         result = job.get_result()
         self._feature_set = result._get_feature_set()
+        self._feature_set_header.feature_set_version = self._feature_set.version
         return result
 
-    def revert_last_ingest(self):
-        request = pb.RevertLastIngestRequest()
+    def materialize_online_async(self):
+        request = pb.StartMaterializationOnlineRequest()
         request.feature_set.CopyFrom(self._feature_set)
-        response = self._stub.RevertLastIngest(request)
-        self._feature_set = response.feature_set
-        return self
+        job_id = self._stub.StartMaterializationOnlineJob(request)
+        return MaterializationOnlineJob(self._stub, job_id)
+
+    @interactive_console.record_stats
+    def materialize_online(self):
+        job = self.materialize_online_async()
+        while not job.is_done():
+            job.show_progress()
+            time.sleep(2)
+
+        job.show_progress()  # there is possibility that some progress was pushed before finishing job
+        result = job.get_result()
+        return result
+
+    def ingest_online(self, rows):
+        if isinstance(rows, list):
+            row_list = rows
+        else:
+            row_list = [rows]
+        request = OnlineApi.OnlineIngestRequest()
+        request.header.CopyFrom(self._feature_set_header)
+        request.rows.extend(row_list)
+        self._stub.OnlineIngest(request)
+
+    def retrieve_online(self, key) -> dict:
+        request = OnlineApi.OnlineRetrieveRequest()
+        request.header.CopyFrom(self._feature_set_header)
+        request.key = str(key)
+        json_row = self._stub.OnlineRetrieve(request).row
+        return json.loads(json_row)
 
     def retrieve(self, start_date_time=None, end_date_time=None):
         return RetrieveHolder(
@@ -380,7 +507,7 @@ class FeatureSet:
         return [VersionDescription(version) for version in response.versions]
 
     def get_schema(self):
-        return Schema.load(self)
+        return Schema.create_from(self)
 
     def ingest_history(self):
         return IngestHistory(self._stub, self._feature_set)
@@ -398,44 +525,82 @@ class VersionDescription:
 
 
 class TimeToLive:
-    def __init__(self, feature_set):
-        self._feature_set = feature_set
-        self._ttl = self._feature_set.time_to_live
+    def __init__(self, stub, feature_set):
+        self._stub = stub
+        self._fs = feature_set
 
     @property
     def ttl_offline(self):
-        return self._ttl.ttl_offline
+        return self._fs._feature_set.time_to_live.ttl_offline
 
     @ttl_offline.setter
     def ttl_offline(self, value):
-        self._ttl.ttl_offline = value
+        update_request = pb.FeatureSetIntFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._fs._feature_set_header)
+        self._stub.UpdateTimeToLiveOfflineInterval(update_request)
+        self._fs.refresh()
 
     @property
     def ttl_offline_interval(self):
-        return self._ttl.ttl_offline_interval
+        return TimeToLiveApi.Offline.TimeToLiveInterval.Name(
+            self._fs._feature_set.time_to_live.ttl_offline_interval
+        )
 
     @ttl_offline_interval.setter
     def ttl_offline_interval(self, value):
-        self._ttl.ttl_offline_interval = value
+        valid_units = TimeToLiveApi.Offline.TimeToLiveInterval.keys()
+        if value.upper() in valid_units:
+            update_request = pb.FeatureSetTimeToLiveOfflineIntervalUnitUpdateRequest()
+            update_request.new_value = TimeToLiveApi.Offline.TimeToLiveInterval.Value(
+                value.upper()
+            )
+            update_request.header.CopyFrom(self._fs._feature_set_header)
+            self._stub.UpdateTimeToLiveOfflineIntervalUnit(update_request)
+            self._fs.refresh()
+        else:
+            raise Exception(
+                "Invalid offline time to live interval unit. Supported values are: "
+                + ", ".join(map(str, valid_units))
+            )
 
     @property
     def ttl_online(self):
-        return self._ttl.ttl_online
+        return self._fs._feature_set.time_to_live.ttl_online
 
     @ttl_online.setter
     def ttl_online(self, value):
-        self._ttl.ttl_online = value
+        update_request = pb.FeatureSetIntFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._fs._feature_set_header)
+        self._stub.UpdateTimeToLiveOnlineInterval(update_request)
+        self._fs.refresh()
 
     @property
     def ttl_online_interval(self):
-        return self._ttl.ttl_online_interval
+        return TimeToLiveApi.Online.TimeToLiveInterval.Name(
+            self._fs._feature_set.time_to_live.ttl_online_interval
+        )
 
     @ttl_online_interval.setter
     def ttl_online_interval(self, value):
-        self._ttl.ttl_online_interval = value
+        valid_units = TimeToLiveApi.Online.TimeToLiveInterval.keys()
+        if value.upper() in valid_units:
+            update_request = pb.FeatureSetTimeToLiveOnlineIntervalUnitUpdateRequest()
+            update_request.new_value = TimeToLiveApi.Online.TimeToLiveInterval.Value(
+                value.upper()
+            )
+            update_request.header.CopyFrom(self._fs._feature_set_header)
+            self._stub.UpdateTimeToLiveOnlineIntervalUnit(update_request)
+            self._fs.refresh()
+        else:
+            raise Exception(
+                "Invalid online time to live interval unit. Supported values are: "
+                + ", ".join(map(str, valid_units))
+            )
 
     def __repr__(self):
-        return Utils.pretty_print_proto(self._ttl)
+        return Utils.pretty_print_proto(self._fs._feature_set.time_to_live)
 
 
 class FeatureSetScope:
@@ -455,49 +620,33 @@ class FeatureSetScope:
         return Utils.pretty_print_proto(self._scope)
 
 
-class SpecialData:
-    def __init__(self, feature_set):
-        self._feature_set = feature_set
-        self._special_data = self._feature_set.special_data
+class FeatureSetSpecialData:
+    def __init__(self, stub, feature_set):
+        self._stub = stub
+        self._fs = feature_set
 
     @property
     def spi(self):
-        return self._special_data.spi
-
-    @spi.setter
-    def spi(self, value):
-        self._special_data.spi = value
+        return self._fs._feature_set.special_data.spi
 
     @property
     def pci(self):
-        return self._special_data.pci
-
-    @pci.setter
-    def pci(self, value):
-        self._special_data.pci = value
+        return self._fs._feature_set.special_data.pci
 
     @property
     def rpi(self):
-        return self._special_data.rpi
-
-    @rpi.setter
-    def rpi(self, value):
-        self._special_data.rpi = value
+        return self._fs._feature_set.special_data.rpi
 
     @property
     def demographic(self):
-        return self._special_data.demographic
-
-    @demographic.setter
-    def demographic(self, value):
-        self._special_data.demographic = value
+        return self._fs._feature_set.special_data.demographic
 
     @property
     def legal(self):
-        return Legal(self._feature_set)
+        return Legal(self._stub, self._fs)
 
     def __repr__(self):
-        return Utils.pretty_print_proto(self._special_data)
+        return Utils.pretty_print_proto(self._fs._feature_set.special_data)
 
 
 class Statistics:
@@ -514,36 +663,42 @@ class Statistics:
 
 
 class Legal:
-    def __init__(self, feature_set):
-        self._feature_set = feature_set
-        self._legal = self._feature_set.special_data.legal
+    def __init__(self, stub, feature_set):
+        self._stub = stub
+        self._fs = feature_set
 
     @property
     def approved(self):
-        return self._legal.approved
+        return self._fs._feature_set.special_data.legal.approved
 
     @approved.setter
     def approved(self, value):
-        self._legal.approved = value
+        update_request = pb.FeatureSetBooleanFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._fs._feature_set_header)
+        self._stub.UpdateFeatureSetLegalApproved(update_request)
+        self._fs.refresh()
 
     @property
     def approved_date(self):
-        return self._legal.approved_date
-
-    @approved_date.setter
-    def approved_date(self, value):
-        self._legal.approved_date = value
+        return Utils.timestamp_to_string(
+            self._fs._feature_set.special_data.legal.approved_date
+        )
 
     @property
     def notes(self):
-        return self._legal.notes
+        return self._fs._feature_set.special_data.legal.notes
 
     @notes.setter
     def notes(self, value):
-        self._legal.notes = value
+        update_request = pb.FeatureSetStringFieldUpdateRequest()
+        update_request.new_value = value
+        update_request.header.CopyFrom(self._fs._feature_set_header)
+        self._stub.UpdateFeatureSetLegalApprovedNotes(update_request)
+        self._fs.refresh()
 
     def __repr__(self):
-        return Utils.pretty_print_proto(self._legal)
+        return Utils.pretty_print_proto(self._fs._feature_set.special_data.legal)
 
 
 class Online:
