@@ -45,4 +45,4 @@ class Progress:
             if response.status:
                 self.next_index += len(response.status)
                 for status in response.status:
-                    log(status)
+                    log(f"Job ID: {self.job.job_id}, Status: {status}")
