@@ -22,6 +22,7 @@ class RetrieveAsLinksCommon:
         while not self._get_job(self._job_id).done:
             progress.show()
             time.sleep(2)
+        progress.show()  # there is possibility that some progress was pushed before finishing job
         retrieve_as_links_response = self._get_retrieve_links_output(self._job_id)
         return RetrieveAsLinksCommon._download_all(
             output_dir, retrieve_as_links_response.download_links
@@ -75,12 +76,7 @@ class RetrieveHolder(RetrieveAsLinksCommon):
             self._job_id = self._start_retrieve_links_job()
         return super(RetrieveHolder, self).download_async(output_dir)
 
-    def preview(self, num_rows=10):
-        if num_rows > 100:
-            print(
-                "Preview can retrieve at most 100 rows. Please try with a lesser value."
-            )
-            return
+    def _start_preview_job(self, num_rows):
         request = pb.PreviewRequest()
         request.feature_set.CopyFrom(self._feature_set)
         if self._start_date_time is not None:
@@ -88,7 +84,25 @@ class RetrieveHolder(RetrieveAsLinksCommon):
         if self._end_date_time is not None:
             request.end_date_time = self._end_date_time
         request.num_rows = num_rows
-        resp = self._stub.GetPreview(request)
+        request.ingest_id = self._ingest_id
+        return self._stub.StartPreviewJob(request)
+
+    def preview(self, num_rows=10):
+        if num_rows > 100:
+            print(
+                "Preview can retrieve at most 100 rows. Please try with a lesser value."
+            )
+            return
+
+        if not self._job_id:
+            self._job_id = self._start_preview_job(num_rows)
+
+        progress = interactive_console.Progress(self._stub, self._job_id)
+        while not self._get_job(self._job_id).done:
+            progress.show()
+            time.sleep(2)
+        progress.show()  # there is possibility that some progress was pushed before finishing job
+        resp = self._stub.GetPreviewJobOutput(self._job_id)
         print(resp.result)
 
     def as_spark_frame(self, spark_session):
@@ -101,9 +115,10 @@ class RetrieveHolder(RetrieveAsLinksCommon):
         )
 
         if self._retrieve_as_spark_response is None:
-            request = self._create_retrieve_request()
+            request = self._create_retrieve_request(spark_session)
             self._retrieve_as_spark_response = self._stub.RetrieveAsSpark(request)
         resp = self._retrieve_as_spark_response
+        spark_session.conf.set("ai.h2o.featurestore.sessionId", resp.session_id)
         SparkUtils.configure_user_spark(spark_session)
         for k, v in resp.options.items():
             spark_session.conf.set(k, v)
@@ -130,16 +145,20 @@ class RetrieveHolder(RetrieveAsLinksCommon):
                 )
                 .filter(col(timestamp_col) <= end_timestamp)
                 .filter(col(timestamp_col) >= start_timestamp)
-            )
+            ).drop(timestamp_col)
         else:
             output_df = df.filter(
                 col("time_travel_column_auto_generated") <= end_timestamp
             ).filter(col("time_travel_column_auto_generated") >= start_timestamp)
         return output_df.drop("ingest_id")
 
-    def _create_retrieve_request(self):
+    def _create_retrieve_request(self, spark_session=None):
+        session_id = ""
+        if spark_session is not None:
+            session_id = spark_session.conf.get("ai.h2o.featurestore.sessionId", "")
         request = pb.RetrieveRequest()
         request.feature_set.CopyFrom(self._feature_set)
+        request.session_id = session_id
         if self._start_date_time is not None:
             request.start_date_time = self._start_date_time
         if self._end_date_time is not None:
